from __future__ import annotations

import asyncio
import socket
from contextlib import closing
from io import BytesIO

import pytest
from pytest_lazy_fixtures import lf

import memcachio
from memcachio import TCPConnection, UnixSocketConnection
from memcachio.commands import Command, FlushAllCommand, GetCommand, R, Request, SetCommand
from memcachio.errors import ClientError, MemcachioConnectionError, ServerError
from memcachio.types import SingleMemcachedInstanceEndpoint
from tests.conftest import flush_server


class TestConnectionErrors:
    @pytest.mark.parametrize("address", [("192.0.2.0", 11211), ("100::1", 11211)])
    async def test_tcp_host_no_response(self, address):
        connection = memcachio.TCPConnection(address, connect_timeout=0.1)
        with closing(connection):
            with pytest.raises(MemcachioConnectionError):
                await connection.connect()

    async def test_tcp_host_wrong_type_of_server(self):
        connection = memcachio.TCPConnection(("8.8.8.8", 53))
        with closing(connection):
            await connection.connect()
            command = GetCommand("test")
            connection.create_request(command)
            assert connection.connected
            with pytest.raises(MemcachioConnectionError):
                await command.response
            assert not connection.connected
            assert connection.metrics.requests_failed == 1

    async def test_uds_invalid_socket(self):
        connection = UnixSocketConnection("/var/tmp/invalid.sock")
        with closing(connection):
            with pytest.raises(MemcachioConnectionError, match="Unable to establish a connection"):
                await connection.connect()

    async def test_connect_timeout(self, memcached_1):
        await flush_server(memcached_1)
        connection = memcachio.TCPConnection(memcached_1, connect_timeout=0.00001)
        with closing(connection):
            with pytest.raises(
                MemcachioConnectionError, match="Unable to establish a connection within"
            ):
                await connection.connect()

    async def test_read_timeout(self, memcached_1):
        await flush_server(memcached_1)
        connection = memcachio.TCPConnection(
            memcached_1, read_timeout=0.0001, max_inflight_requests_per_connection=1024
        )
        with closing(connection):
            await connection.connect()
            set_commands = [SetCommand(f"key{i}", bytes(32 * 1024)) for i in range(10)]
            [connection.create_request(command) for command in set_commands]
            get_commands = [GetCommand(f"key{i}") for i in range(10)]
            [connection.create_request(command) for command in get_commands]
            with pytest.raises(TimeoutError, match="command .* timed out after 0.0001 seconds"):
                await asyncio.gather(*[command.response for command in set_commands + get_commands])

    async def test_server_error(self, memcached_1):
        await flush_server(memcached_1)
        connection = memcachio.TCPConnection(memcached_1)
        with closing(connection):
            await connection.connect()
            flush_all = FlushAllCommand(0)
            connection.create_request(flush_all)
            await flush_all.response
            data = bytes(2 * 1024 * 1024)
            set_command = SetCommand("key", data)
            connection.create_request(set_command)
            with pytest.raises(ServerError, match="object too large for cache"):
                await set_command.response

    async def test_client_error(self, memcached_1):
        await flush_server(memcached_1)
        connection = memcachio.TCPConnection(memcached_1)
        with closing(connection):
            await connection.connect()

            class BadCommand(Command[bool]):
                name = b"set"

                def build_request(self) -> Request[R]:
                    return Request(self, b"key 0 0 2", [b"123\r\n"])

                def parse(self, data: BytesIO, endpoint: SingleMemcachedInstanceEndpoint) -> R:
                    header = data.readline()
                    self._check_header(header)
                    return False

            bad_command = BadCommand()
            connection.create_request(bad_command)
            with pytest.raises(ClientError, match="bad data chunk"):
                await bad_command.response

    @pytest.mark.parametrize(
        "endpoint",
        [pytest.param(lf(target)) for target in ["memcached_1", "memcached_uds"]],
    )
    async def test_abrupt_disconnection(self, endpoint):
        if isinstance(endpoint, tuple):
            connection = TCPConnection(endpoint)
        else:
            connection = UnixSocketConnection(endpoint)
        with closing(connection):
            await connection.connect()
            commands = [SetCommand(f"key{i}", bytes(32 * 1024)) for i in range(4096)]
            asyncio.get_running_loop().call_soon(connection.close)
            [connection.create_request(command) for command in commands]
            responses = await asyncio.gather(
                *[command.response for command in commands], return_exceptions=True
            )
            assert not all([k is True for k in responses])


class TestDataReceived:
    async def test_socket_read_with_newlines(self, memcached_1, mocker):
        await flush_server(memcached_1)
        connection = memcachio.TCPConnection(memcached_1)
        with closing(connection):
            await connection.connect()
            set_command = SetCommand("key", b"\r\n".join([b"this is a", b"multiline sentence"]))
            connection.create_request(set_command)
            assert await set_command.response
            get_command = GetCommand("key")
            connection.create_request(get_command)
            item = await get_command.response
            assert item.get(b"key").value == b"this is a\r\nmultiline sentence"

    async def test_socket_read_batch(self, memcached_1, mocker):
        await flush_server(memcached_1)
        connection = memcachio.TCPConnection(memcached_1)
        with closing(connection):
            await connection.connect()
            set_command = SetCommand("key", bytes(512 * 1024))
            connection.create_request(set_command)
            assert await set_command.response

            data_received = mocker.spy(connection, "data_received")
            get_command = GetCommand("key")
            connection.create_request(get_command)
            item = await get_command.response
            assert item != {}
            assert data_received.call_count > 1


class TestConnectionOptions:
    async def test_socket_no_delay_tcp(self, memcached_1):
        connection = memcachio.TCPConnection(memcached_1, socket_nodelay=False)
        with closing(connection):
            await connection.connect()
            assert 0 == connection._transport.get_extra_info("socket").getsockopt(
                socket.IPPROTO_TCP, socket.TCP_NODELAY
            )

        connection = memcachio.TCPConnection(memcached_1, socket_nodelay=True)
        with closing(connection):
            await connection.connect()
            assert 0 != connection._transport.get_extra_info("socket").getsockopt(
                socket.IPPROTO_TCP, socket.TCP_NODELAY
            )

    async def test_socket_no_delay_uds(self, memcached_uds):
        connection = memcachio.UnixSocketConnection(memcached_uds, socket_nodelay=True)
        with closing(connection):
            await connection.connect()
            assert connection.connected

    async def test_socket_keepalive_options(self, memcached_1):
        connection = memcachio.TCPConnection(
            memcached_1,
            socket_keepalive=False,
        )
        with closing(connection):
            await connection.connect()
            assert 0 == connection._transport.get_extra_info("socket").getsockopt(
                socket.SOL_SOCKET, socket.SO_KEEPALIVE
            )

        connection = memcachio.TCPConnection(
            memcached_1,
            socket_keepalive=True,
            socket_keepalive_options={
                socket.TCP_KEEPINTVL: 1,
                socket.TCP_KEEPCNT: 2,
            },
        )
        with closing(connection):
            await connection.connect()
            assert 0 != connection._transport.get_extra_info("socket").getsockopt(
                socket.SOL_SOCKET, socket.SO_KEEPALIVE
            )
            assert 1 == connection._transport.get_extra_info("socket").getsockopt(
                socket.IPPROTO_TCP, socket.TCP_KEEPINTVL
            )
            assert 2 == connection._transport.get_extra_info("socket").getsockopt(
                socket.IPPROTO_TCP, socket.TCP_KEEPCNT
            )

    async def test_invalid_socket_options(self, memcached_1):
        connection = memcachio.TCPConnection(
            memcached_1,
            socket_keepalive=True,
            socket_keepalive_options={
                666: 1,
            },
        )
        with closing(connection):
            with pytest.raises(MemcachioConnectionError) as exc_info:
                await connection.connect()
            assert "Protocol not available" in exc_info.value.__cause__.args



from __future__ import annotations

import hashlib
import uuid
from collections import Counter

import mmh3
import pytest

from memcachio.errors import NoAvailableNodes
from memcachio.routing import KeyRouter


@pytest.mark.parametrize(
    "hasher",
    [
        None,
        mmh3.hash,
        lambda key: int.from_bytes(hashlib.blake2s(key.encode()).digest(), "big"),
    ],
    ids=["default(md5)", "mmh3", "blake2s"],
)
class TestRouting:
    def test_empty_router(self, hasher):
        router = KeyRouter(set(), hasher)
        with pytest.raises(NoAvailableNodes):
            router.get_node("fubar")

    def test_single_node(self, hasher):
        router = KeyRouter(hasher=hasher)
        router.add_node("/var/tmp/socket")
        assert {"/var/tmp/socket"} == {router.get_node(f"key{i}") for i in range(4096)}

    @pytest.mark.parametrize(
        "nodes",
        [
            ("/var/tmp/1", "/var/tmp/2"),
            (("localhost", 11211), ("localhost", 11212)),
            (("localhost", 11211), ("localhost", 11212), "/var/tmp/socket"),
        ],
    )
    def test_multiple_nodes(self, nodes, hasher):
        router = KeyRouter(set(nodes), hasher=hasher)
        counter = Counter(router.get_node(f"key{i}") for i in range(4096))
        for node in nodes:
            assert 1.0 / len(nodes) == pytest.approx(counter[node] / 4096, 1e-1)

    @pytest.mark.parametrize(
        "nodes",
        [
            ("/var/tmp/1", "/var/tmp/2", "/var/tmp/3"),
            (("localhost", 11211), ("localhost", 11212), ("localhost", 11213)),
        ],
    )
    def test_multiple_nodes_with_removal(self, nodes, hasher):
        router = KeyRouter(set(nodes), hasher=hasher)
        mapping = {}
        for i in range(4096):
            mapping.setdefault(router.get_node(f"key{i}"), set()).add(i)
        router.remove_node(nodes[-1])
        new_mapping = {}
        for i in range(4096):
            new_mapping.setdefault(router.get_node(f"key{i}"), set()).add(i)
        assert mapping[nodes[0]].issubset(new_mapping[nodes[0]])
        assert mapping[nodes[1]].issubset(new_mapping[nodes[1]])

        moved_to_node_0 = mapping[nodes[2]].intersection(new_mapping[nodes[0]])
        moved_to_node_1 = mapping[nodes[2]].intersection(new_mapping[nodes[1]])
        assert pytest.approx(len(moved_to_node_0) / 4096, 1e-1) == pytest.approx(
            len(moved_to_node_1) / 4096, 1e-1
        )

    @pytest.mark.benchmark(group="router")
    def test_hash_performance(self, benchmark, hasher):
        nodes = {("localhost", i) for i in range(10)}
        keys = [uuid.uuid4().hex for _ in range(pow(2, 9))]
        router = KeyRouter(nodes, hasher=hasher)
        benchmark(lambda: (router.get_node(key) for key in keys))



from __future__ import annotations

import asyncio
import os
import platform
import socket
import ssl
import sys
import time

import pytest
from pytest_lazy_fixtures import lf

import memcachio
from memcachio.types import TCPEndpoint


def pypy_flaky_marker():
    return pytest.mark.xfail(sys.implementation.name == "pypy", reason="Flaky on pypy")


def ping_socket(host, port):
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect((host, port))

        if os.environ.get("CI") == "True":
            time.sleep(5)
        return True
    except Exception:
        return False


@pytest.fixture(scope="session")
def host_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(("10.255.255.255", 1))
        ip = s.getsockname()[0]
    except Exception:
        ip = "127.0.0.1"
    finally:
        s.close()

    return ip


@pytest.fixture(scope="session")
def host_ip_env(host_ip):
    os.environ["HOST_IP"] = str(host_ip)


class FakeElasticacheServer:
    def __init__(self, servers: dict[str, tuple[str, int]]):
        self.servers = servers
        self.server = None
        self.version = 1

    @property
    def location(self) -> tuple[str, int]:
        return ("localhost", 55551)

    async def start(self) -> None:
        self.server = await asyncio.start_server(self.handle_client, "localhost", 55551)
        await self.server.serve_forever()

    async def handle_client(self, reader, writer):
        data = await reader.read(100)
        message = data.decode().lower()
        if "config get cluster" in message:
            servers = []
            for server, location in self.servers.items():
                servers.append(f"{server}|{location[0]}|{location[1]}")

            response = f"CONFIG cluster 0 {len(' '.join(servers))}\r\n"
            response += f"{self.version}\n"
            response += f"{' '.join(servers)}\n"
            response += "\r\n"
            response += "END\r\n"
        else:
            response = f"Command '{message}' not recognized\n"

        writer.write(response.encode())
        await writer.drain()
        writer.close()

    def add_server(self, name: str, endpoint: tuple[str, int]):
        self.servers[name] = endpoint
        self.version += 1

    def remove_server(self, name):
        self.servers.pop(name)
        self.version += 1

    async def stop(self):
        if self.server:
            self.server.close()
            await self.server.wait_closed()


@pytest.fixture
async def elasticache_endpoint(memcached_1, memcached_2):
    # Configure the server with the list of memcached server addresses
    server = FakeElasticacheServer(
        servers={"memcached-1": memcached_1, "memcached-2": memcached_2},
    )

    asyncio.ensure_future(server.start())
    await asyncio.sleep(0.1)
    yield server
    await server.stop()


@pytest.fixture(scope="session")
def docker_services(host_ip_env, docker_services):
    return docker_services


@pytest.fixture(scope="session")
def memcached_1(docker_services):
    docker_services.start("memcached-1")
    docker_services.wait_for_service("memcached-1", 11211, ping_socket)
    yield ("localhost", 44441)


@pytest.fixture(scope="session")
def memcached_2(docker_services):
    docker_services.start("memcached-2")
    docker_services.wait_for_service("memcached-2", 11211, ping_socket)
    yield ("localhost", 44442)


@pytest.fixture(scope="session")
def memcached_3(docker_services):
    docker_services.start("memcached-3")
    docker_services.wait_for_service("memcached-3", 11211, ping_socket)
    yield ("localhost", 44443)


@pytest.fixture(scope="session")
def memcached_sasl(docker_services):
    docker_services.start("memcached-sasl")
    docker_services.wait_for_service("memcached-sasl", 11211, ping_socket)
    yield ("localhost", 44445)


@pytest.fixture(scope="session")
def memcached_cluster(memcached_1, memcached_2, memcached_3):
    yield (memcached_1, memcached_2, memcached_3)


@pytest.fixture(scope="session")
def memcached_ssl(docker_services):
    docker_services.start("memcached-ssl")
    docker_services.wait_for_service("memcached-ssl", 11211, ping_socket)
    yield ("localhost", 44444)


@pytest.fixture(scope="session")
def memcached_uds(docker_services):
    if platform.system().lower() == "darwin":
        pytest.skip("Fixture not supported on OSX")
    docker_services.start("memcached-uds")
    yield "/tmp/memcachio.sock"


@pytest.fixture
async def memcached_tcp_client(memcached_1, request):
    client = memcachio.Client(memcached_1)
    await client.flushall()
    yield client
    client.connection_pool.close()


@pytest.fixture
async def memcached_ssl_client(memcached_ssl, request):
    ssl_context = ssl.create_default_context()
    ssl_context.check_hostname = False
    ssl_context.verify_mode = ssl.VerifyMode.CERT_NONE
    client = memcachio.Client(memcached_ssl, ssl_context=ssl_context)
    await client.flushall()
    yield client
    client.connection_pool.close()


@pytest.fixture
async def memcached_uds_client(memcached_uds, request):
    client = memcachio.Client(memcached_uds)
    await client.flushall()
    yield client
    client.connection_pool.close()


@pytest.fixture
async def memcached_tcp_cluster_client(memcached_1, memcached_2, memcached_3, request):
    client = memcachio.Client(
        [TCPEndpoint(*memcached_1), TCPEndpoint(*memcached_2), TCPEndpoint(*memcached_3)]
    )
    await client.flushall()
    yield client
    client.connection_pool.close()


@pytest.fixture(scope="session")
def docker_services_project_name():
    return "memcachio"


@pytest.fixture(scope="session")
def docker_compose_files(pytestconfig):
    """Get the docker-compose.yml absolute path.
    Override this fixture in your tests if you need a custom location.
    """

    return ["docker-compose.yml"]


def targets(*targets):
    return pytest.mark.parametrize(
        "client",
        [pytest.param(lf(target)) for target in targets],
    )


async def flush_server(endpoint):
    client = memcachio.Client(endpoint)
    await client.flushall()


@pytest.fixture(autouse=True)
async def wait_for_async_tasks():
    yield
    pending = {t for t in asyncio.all_tasks() if t is not asyncio.current_task() and not t.done()}
    [t.cancel() for t in pending]
    await asyncio.gather(*pending, return_exceptions=True)



from __future__ import annotations

import asyncio
import re
from unittest.mock import ANY

import pytest

import memcachio
from memcachio.errors import ClientError
from tests.conftest import targets


@targets(
    "memcached_tcp_client",
    "memcached_ssl_client",
    "memcached_tcp_cluster_client",
    "memcached_uds_client",
)
class TestCommands:
    async def test_get(self, client: memcachio.Client):
        assert {} == await client.get("not-exist")
        assert await client.set("exists", 1)
        assert {b"exists": ANY} == await client.get("exists")

    async def test_get_many(self, client: memcachio.Client):
        values = {f"key-{i}": i for i in range(100)}
        assert {} == await client.get(*values.keys())
        for key, value in values.items():
            assert await client.set(key, value)
        assert {key.encode("utf-8"): ANY for key in values.keys()} == await client.get(
            *values.keys(), "no-key"
        )

    async def test_gets(self, client: memcachio.Client):
        assert {} == await client.get("not-exist")
        assert await client.set("exists", 1)
        assert (await client.gets("exists")).get(b"exists").cas is not None

    async def test_gets_many(self, client: memcachio.Client):
        values = {f"key-{i}": i for i in range(100)}
        assert {} == await client.gets(*values.keys())
        for key, value in values.items():
            assert await client.set(key, value)
        cass = [value.cas for _, value in (await client.gets(*values.keys())).items()]
        assert len(cass) == 100
        assert not any(cas is None for cas in cass)

    async def test_gat(self, client: memcachio.Client):
        assert {} == await client.gat("not-exist", expiry=1)
        assert await client.set("exists", 1)
        assert {b"exists": ANY} == await client.gat("exists", expiry=1)
        await asyncio.sleep(1)
        assert {} == await client.get("exists")

    async def test_gat_many(self, client: memcachio.Client):
        values = {f"key-{i}": i for i in range(100)}
        assert {} == await client.gat(*values.keys(), expiry=1)
        for key, value in values.items():
            assert await client.set(key, value)
        assert len((await client.gat(*values.keys(), expiry=1)).items()) == 100
        await asyncio.sleep(1)
        assert len((await client.gat(*values.keys(), expiry=1)).items()) == 0

    async def test_gats(self, client: memcachio.Client):
        assert {} == await client.gats("not-exist", expiry=1)
        assert await client.set("exists", 1)
        assert (await client.gats("exists", expiry=1)).get(b"exists").cas is not None
        assert {} == await client.gats("exist", expiry=1)

    async def test_gats_many(self, client: memcachio.Client):
        values = {f"key-{i}": i for i in range(100)}
        assert {} == await client.gats(*values.keys(), expiry=1)
        for key, value in values.items():
            assert await client.set(key, value)
        assert len((await client.gats(*values.keys(), expiry=1)).items()) == 100
        await asyncio.sleep(1)
        assert len((await client.gats(*values.keys(), expiry=1)).items()) == 0

    async def test_set(self, client: memcachio.Client):
        assert await client.set("key", 1)
        assert (await client.get("key")).get(b"key").value == b"1"

        assert await client.set("key", 2, 1)
        assert (await client.get("key")).get(b"key").flags == 1

        assert await client.set("key", 3, expiry=1)
        await asyncio.sleep(1)
        assert not await client.get("key")

        assert None is await client.set("key", 4, noreply=True)
        assert (await client.get("key")).get(b"key").value == b"4"

    async def test_cas(self, client: memcachio.Client):
        assert await client.set("key", 1)
        cas = (await client.gets("key")).get(b"key").cas
        assert cas is not None
        assert not await client.cas("key", 2, cas + 1)
        assert await client.cas("key", 2, cas)
        item = (await client.gets("key")).get(b"key")
        assert item.cas != cas
        assert item.value == b"2"

    async def test_add(self, client: memcachio.Client):
        assert await client.set("key", 1)
        assert not await client.add("key", 2)
        assert await client.add("newkey", 2)
        item = (await client.get("newkey")).get(b"newkey")
        assert item.value == b"2"

    async def test_append(self, client: memcachio.Client):
        assert not await client.append("key", "o")
        assert await client.set("key", "fo")
        assert await client.append("key", "o")
        item = (await client.get("key")).get(b"key")
        assert item.value == b"foo"

    async def test_prepend(self, client: memcachio.Client):
        assert not await client.prepend("key", "f")
        assert await client.set("key", "oo")
        assert await client.prepend("key", "f")
        item = (await client.get("key")).get(b"key")
        assert item.value == b"foo"

    async def test_replace(self, client: memcachio.Client):
        assert not await client.replace("key", 1)
        assert await client.set("key", 1)
        item = (await client.get("key")).get(b"key")
        assert item.value == b"1"
        assert await client.replace("key", 2, expiry=2)
        item = (await client.get("key")).get(b"key")
        assert item.value == b"2"
        await asyncio.sleep(2)
        assert {} == await client.get("key")

    async def test_delete(self, client: memcachio.Client):
        assert await client.set("key", 1)
        assert not await client.delete("no-key")
        assert await client.delete("key")

    async def test_touch(self, client: memcachio.Client):
        assert await client.set("key", 1)
        assert await client.touch("key", 1)
        await asyncio.sleep(1)
        assert not await client.delete("key")

    async def test_incr(self, client: memcachio.Client):
        assert await client.incr("key", 1) is None
        assert await client.set("key", 1)
        assert 2 == await client.incr("key", 1)
        assert await client.incr("key", 1, noreply=True) is None
        assert 4 == await client.incr("key", 1)

        with pytest.raises(ClientError, match="invalid numeric delta"):
            await client.incr("key", pow(2, 64))

        await client.set("other-key", "one")
        with pytest.raises(ClientError, match="non-numeric value"):
            await client.incr("other-key", 1)

    async def test_decr(self, client: memcachio.Client):
        assert await client.decr("key", 1) is None
        assert await client.set("key", 3)
        assert 2 == await client.decr("key", 1)
        assert await client.decr("key", 1, noreply=True) is None
        assert 0 == await client.decr("key", 1)

        with pytest.raises(ClientError, match="invalid numeric delta"):
            await client.decr("key", pow(2, 64))

        await client.set("other-key", "one")
        with pytest.raises(ClientError, match="non-numeric value"):
            await client.decr("other-key", 1)

    async def test_stats(self, client: memcachio.Client):
        stats = await client.stats()
        for instance, stats in stats.items():
            assert b"bytes_read" in stats
        slab_stats = await client.stats("slabs")
        for instance, stats in slab_stats.items():
            assert b"active_slabs" in stats

    async def test_version(self, client: memcachio.Client):
        versions = await client.version()
        for instance, version in versions.items():
            assert re.match(r"\d+.\d+.\d+", version)

    async def test_all_noreply(self, client: memcachio.Client):
        assert None is await client.set("fubar", 1, noreply=True)
        assert None is await client.cas("fubar", 1, 0, noreply=True)
        assert None is await client.add("fubar", 1, noreply=True)
        assert None is await client.incr("fubar", 1, noreply=True)
        assert None is await client.decr("fubar", 2, noreply=True)
        assert b"0" == (await client.get("fubar")).get(b"fubar").value
        assert None is await client.replace("fubar", 6, noreply=True)
        assert None is await client.append("fubar", 6, noreply=True)
        assert None is await client.prepend("fubar", 6, noreply=True)
        value = await client.gat("fubar", "fubar2", expiry=2)
        assert None is await client.touch("fubar", 1, noreply=True)
        await asyncio.sleep(1)
        assert None is await client.delete("fubar2", noreply=True)
        assert value.get(b"fubar").value == b"666"
        assert {} == await client.get("fubar", "fubar2")



#!/usr/bin/env python3
from __future__ import annotations

import os
import sys

sys.path.insert(0, os.path.abspath("../../"))
sys.path.insert(0, os.path.abspath("./"))

import requests
from theme_config import *

import memcachio
import memcachio.types

try:
    latest_version = requests.get("https://pypi.org/pypi/memcachio/json").json()["info"][
        "version"
    ]
except:
    latest_version = None

master_doc = "index"
project = "memcachio"
copyright = "2025, Ali-Akber Saifee"
author = "alisaifee"
description = "Async memcached client for python"

html_static_path = ["./_static"]
html_css_files = [
    "custom.css",
    "https://fonts.googleapis.com/css2?family=Fira+Code:wght@300;400;700&family=Fira+Sans:ital,wght@0,100;0,200;0,300;0,400;0,500;0,600;0,700;0,800;0,900;1,100;1,200;1,300;1,400;1,500;1,600;1,700;1,800;1,900&display=swap",
]

html_baseurl = "https://memcachio.readthedocs.io/"
sitemap_url_scheme = "en/stable/{link}"

extensions = [
    "enum_tools.autoenum",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.autosummary",
    "sphinx.ext.extlinks",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
    "sphinx_issues",
    "sphinx_sitemap",
    "sphinxcontrib.programoutput",
    "sphinxext.opengraph",
]

autodoc_default_options = {
    "members": True,
    "inherited-members": True,
    "inherit-docstrings": True,
    "member-order": "bysource",
}

ahead = 0

if ".post0.dev" in memcachio.__version__:
    version, ahead = memcachio.__version__.split(".post0.dev")
else:
    version = memcachio.__version__

release = version

html_title = f"{project} <small><b style='color: var(--color-brand-primary)'>{{{release}}}</b></small>"
try:
    ahead = int(ahead)

    if ahead > 0:
        html_theme_options[
            "announcement"
        ] = f"""
        This is a development version. The documentation for the latest version: <b>{latest_version or release}</b> can be found <a href="/en/stable">here</a>
        """
        html_title = f"{project} <small><b style='color: var(--color-brand-primary)'>{{dev}}</b></small>"
except:
    pass

add_module_names = False
autodoc_typehints_format = "short"
autodoc_typehints = "both"
autodoc_preserve_defaults = True
autodoc_type_aliases = {
    "KeyT": "~memcachio.types.KeyT",
    "ValueT": "~memcachio.types.ValueT",
    "UnixSocketLocator": "~memcachio.types.UnixSocketLocator",
}
autosectionlabel_maxdepth = 3
autosectionlabel_prefix_document = True

extlinks = {
    "pypi": ("https://pypi.org/project/%s", "%s"),
}


issues_github_path = "alisaifee/memcachio"


htmlhelp_basename = "memcachiodoc"
latex_elements = {}

latex_documents = [
    (master_doc, "memcachio.tex", "memachio Documentation", "alisaifee", "manual"),
]
man_pages = [(master_doc, "memcachio", "memachio Documentation", [author], 1)]

texinfo_documents = [
    (
        master_doc,
        "memcachio",
        "memcachio Documentation",
        author,
        "memcachio",
        "One line description of project.",
        "Miscellaneous",
    ),
]
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}

# Workaround for https://github.com/sphinx-doc/sphinx/issues/9560
from sphinx.domains.python import PythonDomain
from sphinx.ext.autodoc import ClassDocumenter, Documenter, _

assert PythonDomain.object_types["data"].roles == ("data", "obj")
PythonDomain.object_types["data"].roles = ("data", "class", "obj")

# Workaround for https://github.com/sphinx-doc/sphinx/issues/10333
from sphinx.util import inspect

inspect.TypeAliasForwardRef.__repr__ = lambda self: self.name
inspect.TypeAliasForwardRef.__hash__ = lambda self: hash(self.name)

ClassDocumenter.get_overloaded_signatures = lambda *_: []




from __future__ import annotations

import asyncio
import ssl
from contextlib import closing

import pytest

import memcachio
from memcachio import Authenticator, AWSAutoDiscoveryEndpoint, BaseConnection
from memcachio.commands import Command, SetCommand
from memcachio.errors import ClientError, MemcachioConnectionError
from memcachio.pool import ClusterPool, Pool, PoolMetrics, R, SingleServerPool
from memcachio.types import TCPEndpoint


class TestClient:
    async def test_invalid_construction(self, mocker):
        with pytest.raises(ValueError, match="One of `memcached_location` or `connection_pool`"):
            memcachio.Client(None)
        with pytest.raises(
            ValueError, match="One of `memcached_location` or `connection_pool`.*not both"
        ):
            memcachio.Client("fubar", connection_pool=mocker.Mock())

    async def test_construction_with_single_tcp_endpoint(self, memcached_1):
        client = memcachio.Client(TCPEndpoint(*memcached_1))
        assert isinstance(client.connection_pool, SingleServerPool)

    async def test_construction_with_multiple_tcp_endpoints(self, memcached_1, memcached_2):
        client = memcachio.Client([TCPEndpoint(*memcached_1), TCPEndpoint(*memcached_2)])
        assert isinstance(client.connection_pool, ClusterPool)

    async def test_ssl_context(self, memcached_ssl):
        ssl_context = ssl.create_default_context()
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.VerifyMode.CERT_REQUIRED
        client = memcachio.Client(TCPEndpoint(*memcached_ssl), ssl_context=ssl_context)
        with pytest.raises(MemcachioConnectionError):
            await client.version()
        ssl_context.verify_mode = ssl.VerifyMode.CERT_NONE
        assert await client.version()

    async def test_client_from_custom_pool(self, memcached_1):
        class MyPool(Pool):
            async def initialize(self) -> None:
                pass

            def close(self) -> None:
                pass

            @property
            def metrics(self) -> PoolMetrics:
                return PoolMetrics()

            async def execute_command(self, command: Command[R]) -> None:
                command.response.set_result(True)

        client = memcachio.Client(connection_pool=MyPool(memcached_1))
        assert await client.set("fubar", 1)

    async def test_sasl_authentication(self, memcached_sasl):
        client = memcachio.Client(memcached_sasl)
        with closing(client.connection_pool):
            with pytest.raises(ClientError, match="unauthenticated"):
                await client.get("test")
            client = memcachio.Client(memcached_sasl, username="user", password="wrong")
            with pytest.raises(ClientError, match="authentication failure"):
                await client.get("test")
            client = memcachio.Client(memcached_sasl, username="user", password="password")
            await client.get("test")

    async def test_sasl_authentication_with_custom_authenticator(self, memcached_sasl):
        class MyAuthenticator(Authenticator):
            def __init__(self, username: str, password: str, lie: bool = False):
                self.username = username
                self.password = password
                self.lie = lie

            async def authenticate(self, connection: BaseConnection) -> bool:
                value = f"{self.username} {self.password}" if not self.lie else "lies lies"
                command = SetCommand("auth", value)
                connection.create_request(command)
                return await command.response

        client = memcachio.Client(memcached_sasl, authenticator=MyAuthenticator("user", "password"))
        with closing(client.connection_pool):
            await client.get("test")
        lie_client = memcachio.Client(
            memcached_sasl, authenticator=MyAuthenticator("user", "password", True)
        )
        with closing(lie_client.connection_pool):
            with pytest.raises(ClientError, match="authentication failure"):
                await lie_client.get("test")

    async def test_autodiscovery_client(self, elasticache_endpoint, memcached_1):
        client = memcachio.Client(AWSAutoDiscoveryEndpoint(*elasticache_endpoint.location, 0.1))
        assert len(await client.version()) == 2
        elasticache_endpoint.remove_server("memcached-1")
        await asyncio.sleep(0.5)
        assert len(await client.version()) == 1
        elasticache_endpoint.add_server("memcached-1", memcached_1)
        await asyncio.sleep(0.5)
        assert len(await client.version()) == 2



from __future__ import annotations

import asyncio
import uuid
from contextlib import closing

import pytest
from pytest_lazy_fixtures import lf

import memcachio
from memcachio.pool import SingleServerPool
from tests.conftest import pypy_flaky_marker

pytestmark = pypy_flaky_marker()


@pytest.mark.parametrize(
    "endpoint",
    [pytest.param(lf(target)) for target in ["memcached_1", "memcached_cluster"]],
)
class TestConcurrency:
    async def test_single_concurrent_command(self, endpoint):
        client = memcachio.Client(endpoint)
        with closing(client.connection_pool):
            await client.set("key", 0)
            assert list(range(1, 1025)) == (
                sorted(await asyncio.gather(*[client.incr("key", 1) for _ in range(1024)]))
            )

    @pytest.mark.parametrize("max_connections", (1, 2, 4, 32))
    async def test_max_connection(self, endpoint, max_connections, mocker):
        client = memcachio.Client(endpoint, max_connections=max_connections)
        with closing(client.connection_pool):
            total_max_connections = (
                max_connections
                if isinstance(client.connection_pool, SingleServerPool)
                else max_connections * len(client.connection_pool.endpoints)
            )
            await asyncio.gather(*[client.set(f"key{i}", i) for i in range(4096)])
            if isinstance(client.connection_pool, SingleServerPool):
                assert len(client.connection_pool._active_connections) <= total_max_connections
            else:
                assert (
                    sum(
                        len(pool._active_connections)
                        for pool in client.connection_pool._cluster_pools.values()
                    )
                    <= total_max_connections
                )

    @pytest.mark.parametrize("concurrent_requests", (pow(2, 10), pow(2, 11), pow(2, 12)))
    @pytest.mark.parametrize(
        "key_size",
        (1 * 1024, 16 * 1024, 32 * 1024),
    )
    async def test_multiple_commands(self, endpoint, concurrent_requests, key_size):
        client = memcachio.Client(endpoint, max_connections=32)
        with closing(client.connection_pool):
            await client.flushall()

            async def tester(client, request_id):
                value = (uuid.uuid4().hex.encode("utf-8") + bytes(key_size))[:key_size]
                assert await client.set(f"{request_id}", value)
                return (await client.get(f"{request_id}")).get(
                    f"{request_id}".encode()
                ).value == value

            assert {True} == set(
                await asyncio.gather(*[tester(client, i) for i in range(concurrent_requests)])
            )



from __future__ import annotations

import asyncio
import re
from contextlib import closing
from unittest.mock import ANY, call

import pytest
from pytest_lazy_fixtures import lf

from memcachio import BaseConnection, ClusterPool, SingleServerPool
from memcachio.commands import FlushAllCommand, GetCommand, SetCommand, TouchCommand, VersionCommand
from memcachio.errors import ConnectionNotAvailable, MemcachioConnectionError, NoAvailableNodes
from memcachio.pool import EndpointHealthcheckConfig, EndpointStatus
from tests.conftest import pypy_flaky_marker

pytestmark = pypy_flaky_marker()


@pytest.mark.parametrize(
    "endpoint",
    [pytest.param(lf(target)) for target in ["memcached_1", "memcached_uds"]],
)
class TestSingleInstancePool:
    async def test_pool_expansion(self, endpoint):
        pool = SingleServerPool(
            endpoint,
            max_connections=4,
            max_inflight_requests_per_connection=0,
        )
        with closing(pool):
            commands = [VersionCommand() for _ in range(16)]
            await asyncio.gather(*[pool.execute_command(command) for command in commands])
            assert len(pool._active_connections) == 4

    async def test_blocking_timeout(self, endpoint):
        pool = SingleServerPool(
            endpoint,
            max_connections=1,
            blocking_timeout=0.001,
            max_inflight_requests_per_connection=0,
        )
        with closing(pool):
            await pool.execute_command(SetCommand("key", bytes(4096), noreply=True))

            with pytest.raises(
                ConnectionNotAvailable, match="Unable to get a connection.*in 0.001 seconds"
            ):
                await asyncio.gather(*[pool.execute_command(GetCommand("key")) for i in range(16)])

    async def test_idle_connection_timeout(self, endpoint):
        pool = SingleServerPool(
            endpoint,
            max_connections=10,
            min_connections=4,
            max_inflight_requests_per_connection=0,
            idle_connection_timeout=0.5,
        )
        with closing(pool):
            set_command = SetCommand("key", bytes(1024))
            await pool.execute_command(set_command)
            await set_command.response

            gets = [GetCommand("key") for _ in range(1024)]
            await asyncio.gather(*[pool.execute_command(get_command) for get_command in gets])
            await asyncio.gather(*[get_command.response for get_command in gets])

            assert len(pool._active_connections) == 10
            await asyncio.sleep(1.5)
            assert len(pool._active_connections) == 4


@pytest.mark.parametrize(
    "cluster_endpoint",
    [
        pytest.param([lf(t) for t in target])
        for target in [
            ["memcached_1"],
            ["memcached_1", "memcached_2"],
            ["memcached_3", "memcached_uds"],
        ]
    ],
)
class TestClusterPool:
    async def test_cluster_pool_single_key_command(self, cluster_endpoint, mocker):
        pool = ClusterPool(cluster_endpoint)
        with closing(pool):
            command = TouchCommand("key", expiry=1)
            send = mocker.spy(BaseConnection, "send")
            await pool.execute_command(command)

            send.assert_called_once_with(ANY, b"touch key 1\r\n")

    async def test_cluster_pool_multi_key_command(self, cluster_endpoint):
        pool = ClusterPool(cluster_endpoint)
        with closing(pool):
            await asyncio.gather(
                *[pool.execute_command(SetCommand(f"key{i}", i, noreply=True)) for i in range(1024)]
            )
            get_command = GetCommand(*[f"key{i}" for i in range(1024)])
            await pool.execute_command(get_command)
            assert set(range(1024)) == set(
                [int(k.value) for k in (await get_command.response).values()]
            )

    async def test_cluster_pool_keyless_command(self, cluster_endpoint, mocker):
        pool = ClusterPool(cluster_endpoint)
        with closing(pool):
            command = FlushAllCommand(0)
            send = mocker.spy(BaseConnection, "send")
            await pool.execute_command(command)
            send.assert_has_calls(
                [
                    call(ANY, b"flush_all 0\r\n"),
                ]
                * len(cluster_endpoint)
            )
            assert await command.response

    async def test_endpoint_removal(self, cluster_endpoint, mocker):
        pool = ClusterPool(cluster_endpoint)
        endpoints = pool.endpoints
        first_endpoint = list(endpoints).pop()
        with closing(pool):
            pool.remove_endpoint("/var/tmp/not-in-pool")
            assert pool.endpoints == endpoints
            assert first_endpoint in pool.endpoints
            pool.remove_endpoint(first_endpoint)
            assert first_endpoint not in pool.endpoints

    async def test_all_endpoints_removed(self, cluster_endpoint, mocker):
        pool = ClusterPool(cluster_endpoint)
        endpoints = pool.endpoints
        with closing(pool):
            await pool.initialize()
            [pool.remove_endpoint(endpoint) for endpoint in endpoints]
            get = GetCommand("key")
            with pytest.raises(NoAvailableNodes):
                await pool.execute_command(get)

    async def test_endpoint_addition(self, cluster_endpoint, mocker):
        pool = ClusterPool(cluster_endpoint)
        endpoints = pool.endpoints
        new_endpoint = "/var/tmp/new-instance-not-real.sock"
        with closing(pool):
            pool.add_endpoint(new_endpoint)
            assert pool.endpoints == endpoints | {new_endpoint}
            [pool.remove_endpoint(k) for k in endpoints if k != new_endpoint]
            get = GetCommand("key")
            with pytest.raises(MemcachioConnectionError, match=f"memcached server: {new_endpoint}"):
                await pool.execute_command(get)

    async def test_mark_endpoint_unhealthy(self, cluster_endpoint, mocker):
        pool = ClusterPool(cluster_endpoint)
        bad_endpoint = "/var/tmp/new-instance-not-real.sock"
        pool.add_endpoint(bad_endpoint)
        with closing(pool):
            get = GetCommand("key")
            with pytest.raises(MemcachioConnectionError, match=f"memcached server: {bad_endpoint}"):
                await pool.execute_command(get)
            pool.update_endpoint_status(bad_endpoint, EndpointStatus.DOWN)
            get = GetCommand("key")
            await pool.execute_command(get)
            assert {} == await get.response

    async def test_auto_removal_unhealthy_endpoint(self, cluster_endpoint, mocker):
        pool = ClusterPool(
            cluster_endpoint,
            endpoint_healthcheck_config=EndpointHealthcheckConfig(
                remove_unhealthy_endpoints=True, maximum_error_count_for_removal=1
            ),
        )
        with closing(pool):
            await pool.initialize()
            target_endpoint = pool._router.get_node("key")

            async def raise_error(*args):
                exc = MemcachioConnectionError(endpoint=target_endpoint, message="something bad")
                if instance_pool := pool._cluster_pools.get(target_endpoint, None):
                    for connection in instance_pool._active_connections:
                        connection.close()
                    instance_pool.close()
                    instance_pool.metrics.on_connection_error(None, exc)
                raise exc

            with closing(pool):
                get = GetCommand("key")
                mocker.patch.object(
                    pool._cluster_pools[target_endpoint],
                    "execute_command",
                    side_effect=raise_error,
                    autospec=True,
                )
                mocker.patch.object(
                    pool._cluster_pools[target_endpoint],
                    "initialize",
                    side_effect=raise_error,
                    autospec=True,
                )
                with pytest.raises(
                    MemcachioConnectionError,
                    match=re.escape(f"something bad (memcached server: {target_endpoint})"),
                ):
                    await pool.execute_command(get)
                await asyncio.sleep(0.01)
                assert target_endpoint not in pool.endpoints

    async def test_auto_recovery_unhealthy_endpoint(self, cluster_endpoint, mocker):
        pool = ClusterPool(
            cluster_endpoint,
            endpoint_healthcheck_config=EndpointHealthcheckConfig(
                remove_unhealthy_endpoints=True,
                monitor_unhealthy_endpoints=True,
                maximum_error_count_for_removal=1,
            ),
        )
        with closing(pool):
            await pool.initialize()
            target_endpoint = pool._router.get_node("key")

            async def raise_error(*args, **kwargs):
                exc = MemcachioConnectionError(endpoint=target_endpoint, message="something bad")
                instance_pool = pool._cluster_pools.get(target_endpoint, None)
                for connection in instance_pool._active_connections:
                    connection.close()
                instance_pool.metrics.on_connection_error(None, exc)
                raise exc

            with closing(pool):
                get = GetCommand("key")

                mocker.patch.object(
                    pool._cluster_pools[target_endpoint],
                    "execute_command",
                    side_effect=raise_error,
                    autospec=True,
                )
                mocker.patch.object(
                    pool._cluster_pools[target_endpoint],
                    "initialize",
                    side_effect=raise_error,
                    autospec=True,
                )

                with pytest.raises(
                    MemcachioConnectionError,
                    match=re.escape(f"something bad (memcached server: {target_endpoint})"),
                ):
                    await pool.execute_command(get)

                await asyncio.sleep(0.01)
                assert target_endpoint not in pool.endpoints

                mocker.stopall()

                await asyncio.sleep(5)
                assert target_endpoint in pool.endpoints

                get = GetCommand("key")
                await pool.execute_command(get)
                assert {} == await get.response

    async def test_auto_recovery_fail(self, cluster_endpoint, mocker):
        pool = ClusterPool(
            cluster_endpoint,
            endpoint_healthcheck_config=EndpointHealthcheckConfig(
                remove_unhealthy_endpoints=True,
                monitor_unhealthy_endpoints=True,
                maximum_recovery_attempts=2,
                maximum_error_count_for_removal=1,
            ),
        )
        with closing(pool):
            await pool.initialize()
            target_endpoint = pool._router.get_node("key")

            async def raise_error(*args, **kwargs):
                exc = MemcachioConnectionError(endpoint=target_endpoint, message="something bad")
                instance_pool = pool._cluster_pools.get(target_endpoint, None)
                for connection in instance_pool._active_connections:
                    connection.close()
                instance_pool.metrics.on_connection_error(None, exc)
                raise exc

            with closing(pool):
                get = GetCommand("key")

                mocker.patch.object(
                    pool._cluster_pools[target_endpoint],
                    "execute_command",
                    side_effect=raise_error,
                    autospec=True,
                )
                mocker.patch.object(
                    pool._cluster_pools[target_endpoint],
                    "initialize",
                    side_effect=raise_error,
                    autospec=True,
                )

                with pytest.raises(
                    MemcachioConnectionError,
                    match=re.escape(f"something bad (memcached server: {target_endpoint})"),
                ):
                    await pool.execute_command(get)

                await asyncio.sleep(0.01)
                assert target_endpoint not in pool.endpoints
                await asyncio.sleep(2)
                assert target_endpoint not in pool.endpoints
